﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using System.IO;
using System.Runtime.InteropServices;

namespace amodi
{
    public partial class AutoMODIMainForm : Form
    {

		const int MARGIN_UNDER_CLICK = 2;	// 次の行までの空白

        const int WM_COPYDATA = 0x4A;
        const int WM_APP = 0x8000;
        const int WM_AMODI = (WM_APP + 0x400);
        const int WM_EXEC_OCR = (WM_APP + 0x410);

        // WM_COPYDATA command
        const int WMCD_EXISTCHECK = 0x4000;
		const int WMCD_SETPOINT = 0x4001;
		const int WMCD_RESTORE_WINDOW = 0x4002;


        //[StructLayout(LayoutKind.Sequential)]
        private struct COPYDATASTRUCT
        {
            public IntPtr dwData;
            public int cbData;
            public IntPtr lpData;
        }

        public struct COPYDATASTRUCT_send {
            public Int32 dwData;        //送信する32ビット値
            public Int32 cbData;　　　　//lpDataのバイト数
            public string lpData;　　 //送信するデータへのポインタ(0も可能)
        }

        [DllImport("user32.dll", SetLastError = true)]
        private static extern int PostMessage(IntPtr hWnd, uint msg, IntPtr wParam, IntPtr lParam);
        [DllImport("user32.dll", SetLastError = true)]
        private static extern int SendMessage(IntPtr hWnd, uint msg, IntPtr wParam, ref COPYDATASTRUCT_send lparam);
        [DllImport("User32.dll", EntryPoint = "FindWindow")]
        public static extern Int32 FindWindow(String lpClassName, String lpWindowName);

        MODI.Document md;
        System.IO.FileSystemWatcher fsw;
        Queue<string> FileNameQue;

		// parser output //
		int CurLoc;	// mouse cursor position in text.

        public AutoMODIMainForm()
        {
			InitializeComponent();
            md = new MODI.Document();
            FileNameQue = new Queue<string>();

#if true
            string exepath = System.IO.Path.GetTempPath();
            exepath += "amodi\\";

			try
			{
				System.IO.Directory.Delete(exepath);
			} catch { }
            System.IO.Directory.CreateDirectory(exepath);
#else
            string exepath = Application.StartupPath;
            lbFileName.Text = "Watching... - " + exepath;
#endif
            fsw = new System.IO.FileSystemWatcher();
            fsw.Path = exepath;
            fsw.IncludeSubdirectories = false;
            fsw.SynchronizingObject = this;
            fsw.Changed += new System.IO.FileSystemEventHandler(watcher_Changed);
            fsw.EnableRaisingEvents = true;
        }
        void Cleanup()
        {
            fsw.EnableRaisingEvents = false;
            fsw.Dispose();
            fsw= null;
        }

        private void btnDoOCR_Click(object sender, EventArgs e)
        {
            DoOCR(tbFileName.Text);
        }

        private bool DoingOCR = false;
        bool DoOCR(string filename)
        {
            if (DoingOCR) return false;
            DoingOCR = true;

            bool capture_page = miCapturePage.Checked;

			ParseFileName(filename);

            lbStatus.Text = "Recognizing... " + filename;
            tbText.Text = "";
            tbInfo.Text = "";
            for (int i = 0; i < 10; i++){
                try {
                    md.Create(filename);
                } catch {
                    tbInfo.AppendText("MODI Create failure: " + filename + "\r\n");
                    DBW("MODI Create file failure: " + filename);
                    System.Threading.Thread.Sleep(30);
                    continue;
                }
                break;
            }
            DBW("DoOCR:OCR");
            try {
				if (miUseDefLang.Checked){
					md.OCR(MODI.MiLANGUAGES.miLANG_SYSDEFAULT, true, true);
				} else {
					md.OCR(MODI.MiLANGUAGES.miLANG_ENGLISH, true, true);
				}
            } catch {
                DBW("MODI OCR Error: " + filename);
                md.Close();
                DBW("DoOCR:Closed");
                lbStatus.Text = "OCR Error: " + filename;
                DoingOCR = false;
                return false;
            }
            DBW("DoOCR:Completed");
            const int UnderGap = MARGIN_UNDER_CLICK; // 単語のある領域より少し下のpointでも検索対象と認識する空白部分(Y方向)
            int last_x = 0;
            int lineno = 0;
			CurLoc = 0;
            lbPoint.Text = "" + CursorPoint.X + "," + CursorPoint.Y;
            tbInfo.AppendText("Page:" + md.Images.Count + " pt:" + CursorPoint.X + "," + CursorPoint.Y + "\r\n");
            for (int i = 0; i < md.Images.Count; i++)
            {
                MODI.Image image = (MODI.Image)md.Images[i];
                MODI.Layout layout = image.Layout;
				bool outok = capture_page;
				string prevWord = "";
				string prevWord2 = "";
				for (int j = 0; j < layout.Words.Count; j++)
                {
                    MODI.Word word = (MODI.Word)layout.Words[j];
                    bool cr = false;
                    bool curLocSet = false;

                    for (int k = 0; k < word.Rects.Count; k++)
                    {
                        MODI.MiRect rc = (MODI.MiRect)word.Rects[k];
                        int h = rc.Bottom - rc.Top;
                        int w = rc.Right - rc.Left;

                        //bool incursor;
						//tbText.AppendText(word.Text + " (" + rc.Left + "," + rc.Top+ ")\r\n");
                        if (CursorPoint.Y >= rc.Top && CursorPoint.Y <= rc.Bottom+UnderGap){
                            //incursor = true;
							if (!outok)
							{
                                if (CursorPoint.X < rc.Left	// cursorを飛び越えた
                                	|| (CursorPoint.X >= rc.Left && CursorPoint.X <= rc.Right)	// cursorが矩形内
                                	){
                                    outok = true;
									curLocSet = true;
	                                if (prevWord.Length != 0) {
										if (NumPrevWords >= 2 && prevWord2.Length != 0){
											tbText.AppendText(prevWord2 + " " + prevWord + " ");
										} else {
		                                    tbText.AppendText(prevWord + " ");
		                                }
	                                    CurLoc = prevWord.Length + 1;
	                                }
                                } else {
									prevWord2 = prevWord;
									prevWord = word.Text;
								}
							}
                        } else {
                            //incursor = false;
                        }

                        if (outok){
                            tbInfo.AppendText(word.Text + " (" + w + "x" + h + ":" + rc.Left + "," + rc.Top + ")\r\n");
                        }

                        if (last_x > rc.Left){
                            if (tbText.Text!="")
                                cr = true;
                            lineno++;
                        }
                        last_x = rc.Left;
                    }

                    if (outok){
                        if (cr){
                            tbText.AppendText("\r\n");
                            if (curLocSet)
                            	CurLoc += 2;	// CR+LF
                        }

                        tbText.AppendText(word.Text + " ");
                    }
                }
            }
            md.Close(false);
            lbStatus.Text = "Done. " + filename;
            DoingOCR = false;
            return true;
        }
		private void ParseFileName(string filename)
		{
			Regex r = new Regex(@"\((?<x>\d+),(?<y>\d+)\)");
			Match m = r.Match(filename);
			if (m.Success)
			{
				CursorPoint.X = int.Parse(m.Groups["x"].Value);
				CursorPoint.Y = int.Parse(m.Groups["y"].Value);
			}
			r = new Regex(@"-n(?<n>\d+)");
			m = r.Match(filename);
			if (m.Success)
			{
				NumPrevWords = int.Parse(m.Groups["n"].Value);
			}
		}

        //イベントハンドラ
        private void watcher_Changed(System.Object source,
            System.IO.FileSystemEventArgs e)
        {
            switch (e.ChangeType)
            {
                case System.IO.WatcherChangeTypes.Changed:
                    string ext = System.IO.Path.GetExtension(e.FullPath);
                    if (ext==".gif" || ext==".jpg" || ext==".png" || ext==".tif" || ext==".bmp" || ext==".dib")
                    {
                        PostOCR(e.FullPath);
                    }
                    break;
//                case System.IO.WatcherChangeTypes.Created:
                    //break;
//                case System.IO.WatcherChangeTypes.Deleted:
  //                  break;
            }
        }

        private void PostOCR(string filename)
        {
            FileNameQue.Enqueue(filename);
            PostMessage(this.Handle, WM_EXEC_OCR, IntPtr.Zero, IntPtr.Zero);
        }

        private void ExecOCR(string filename)
        {
            if (DoOCR(filename)) {
                string textname = filename + ".txt";
                for (int i = 0; i < 10; i++) {
                    try {
                        StreamWriter writer = new StreamWriter(textname, false, System.Text.Encoding.GetEncoding("utf-16"));
                        writer.WriteLine(CurLoc.ToString());
                        writer.Write(tbText.Text);
                        writer.Close();
                    } catch {
                        tbInfo.AppendText("Write Error:" + textname + "\r\n");
                        System.Threading.Thread.Sleep(30);
                        continue;
                    }
                    break;
                }
            }
            if (!miDebugMode.Checked) {
                try {
                    System.IO.File.Delete(filename);
                }
                catch
                {
                    // 連続して.bmpファイルが作られているため
                    tbInfo.AppendText("Delete Error:" + filename + "\r\n");
                }
            }
        }

        private void miExit_Click(object sender, EventArgs e)
        {
            Close();
        }

        enum AMODI_CMD
        {
            QUERY = 0,
            PAGE_CAPTURE = 1,
        };
        struct POINT
        {
            public int x;
            public int y;
        };
        Point CursorPoint;
        unsafe static Point int2point(IntPtr param)
        {
			POINT pt = (POINT)Marshal.PtrToStructure((IntPtr)param, typeof(POINT));
            Point p = new Point();
            p.X = pt.x;
            p.Y = pt.y;
            return p;
        }
        int NumPrevWords = 1;
        // interface for external app
        protected override void WndProc(ref Message m)
        {
            if (m.Msg == WM_AMODI)
            {
                switch ((AMODI_CMD)m.WParam)
                {
                    case AMODI_CMD.QUERY:
                        m.Result = (IntPtr)m.LParam;
                        return;
                    case AMODI_CMD.PAGE_CAPTURE:
                        m.Result = (IntPtr)(miCapturePage.Checked ? 1 : 0);
                        return;
                    default:
                        break;
                }
            } else
            if (m.Msg == WM_EXEC_OCR) {
                string filename = "";
                while (FileNameQue.Count()!=0) {
                    filename = FileNameQue.Dequeue();
                }
                if (filename != "") {
                    ExecOCR(filename);
                }
            } else
            if (m.Msg == WM_COPYDATA)
            {
                COPYDATASTRUCT cds = new COPYDATASTRUCT();
                cds = (COPYDATASTRUCT)Marshal.PtrToStructure(m.LParam, typeof(COPYDATASTRUCT));
                if (cds.dwData == (IntPtr)WMCD_EXISTCHECK)
                {
                    m.Result = (IntPtr)1;
                    return;
                } else
				if (cds.dwData == (IntPtr)WMCD_SETPOINT){
					CursorPoint = int2point(cds.lpData);
					Console.WriteLine("Cursor:" + CursorPoint.X + "," + CursorPoint.Y);
					return;
				} else
				if (cds.dwData == (IntPtr)WMCD_RESTORE_WINDOW)
				{
					this.Show();
				}
#if false   // 参考
                if (cds.cbData > 0)
                {
                    byte[] data = new byte[cds.cbData];
                    Marshal.Copy(cds.lpData, data, 0, cds.cbData);
                    //あとは data に COPYDATA の内容が入ってるので適宜処理
                }
                m.Result = (IntPtr)1;
#endif
            }
            base.WndProc(ref m);
        }

        private void miCapturePage_Click(object sender, EventArgs e)
        {
            miCapturePage.Checked = !miCapturePage.Checked;
        }

		private void AutoMODIMainForm_Activated(object sender, EventArgs e)
		{
			//this.Hide();
		}

		private void miDebugMode_Click(object sender, EventArgs e)
		{
			miDebugMode.Checked = !miDebugMode.Checked;
		}

		private void miUseDefLang_Click(object sender, EventArgs e)
		{
			miUseDefLang.Checked = !miUseDefLang.Checked;
		}

        static int hWin = 0;
        void DBW( string msg )
        {
	        if ( hWin==0 ){
		        hWin = FindWindow( "TDbgMsgForm", "Debug Messenger" );
		        if ( hWin==0 ) return;
	        }
            byte[] bytearray = System.Text.Encoding.Default.GetBytes(msg);
            int len = bytearray.Length;
	        COPYDATASTRUCT_send cds;
	        cds.dwData = 1;	// Indicate String
	        cds.cbData = len + 1;
            cds.lpData = msg;
	        SendMessage( (IntPtr)hWin, WM_COPYDATA, IntPtr.Zero, ref cds );
        }
    }
}
